-------------------------------------------------------------------------------
-- Title      : Immediate pulse train generator
-- Project    : White Rabbit Network Interface
-------------------------------------------------------------------------------
-- File       : imm_pulse_train_gen.vhd
-- Author     : Jorge Machado
-- Company    : Seven Solutions
-- Created    : 2020-03-19
-- Last update: 
-- Platform   : FPGA-generic
-- Standard   : VHDL
-------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity imm_pulse_train_gen is
  generic(
    pulse_period_width : integer := 28
  );
  port(
      clk_ref_i             : in std_logic;
      rst_n_i               : in std_logic;
      dio_pulse_immed_stb_i : in std_logic;
      pulse_period_i        : in std_logic_vector(pulse_period_width-1 downto 0);
      pulse_output_o        : out std_logic
    );
end imm_pulse_train_gen;

architecture Behavioral of imm_pulse_train_gen is
  
  signal nozeroperiod, nozeroperiod_aux	: boolean;

  signal pulse_period : std_logic_vector(pulse_period_width-1 downto 0);
  
  signal dio_pulse_immed_stb_d0, dio_pulse_immed_stb_d1, 
    dio_pulse_immed_stb_d2, dio_pulse_immed_stb_d3 : std_logic;
  
  -- Internal registers to hold pulse duration
  signal counter : unsigned (pulse_period_width-1 downto 0);

    -- Signals for states
   type counter_state is (WAIT_ST, COUNTING, CAPTURE_PERIOD, TRIGGER);
  signal state : counter_state;
  
  signal repeat_pulse : std_logic;
  
  -- Aux
  constant zeros : std_logic_vector(pulse_period_width-1 downto 0) := (others => '0');
  
  constant initial_pulse_delay_compensation : integer := 7;
  constant repeat_pulse_delay_compensation : integer := 4;

begin

  synchronization : process(clk_ref_i, rst_n_i)
  begin
  if (rst_n_i = '0') then
    dio_pulse_immed_stb_d0 <= '0';
    dio_pulse_immed_stb_d1 <= '0';
    dio_pulse_immed_stb_d2 <= '0';
    dio_pulse_immed_stb_d3 <= '0';
   elsif rising_edge(clk_ref_i) then
    dio_pulse_immed_stb_d0 	<= dio_pulse_immed_stb_i;
    dio_pulse_immed_stb_d1  <= dio_pulse_immed_stb_d0;
    dio_pulse_immed_stb_d2  <= dio_pulse_immed_stb_d1;
    dio_pulse_immed_stb_d3  <= dio_pulse_immed_stb_d2;
    --pulse_period						<= pulse_period_i;			
    nozeroperiod_aux 				<= pulse_period_i /= zeros;
    if ((dio_pulse_immed_stb_d2 = '1' and dio_pulse_immed_stb_d1 = '0') or (state = CAPTURE_PERIOD)) then
      nozeroperiod <= nozeroperiod_aux;
      pulse_period <= pulse_period_i;	
    end if;
  end if;
  end process;
  
  state_process : process(clk_ref_i, rst_n_i)
   begin
    if (rst_n_i = '0') then 
      counter 			<= (others => '0');
      state   			<= WAIT_ST;
      repeat_pulse 	<= '0';
    elsif rising_edge(clk_ref_i) then
      case state is
        when WAIT_ST =>
        if dio_pulse_immed_stb_d3 = '1' and nozeroperiod then
          state   <= COUNTING;
          --Store the period two cycle before than the immed_pulse_counter_process to not lost one cycle.
          counter <= unsigned(pulse_period)-initial_pulse_delay_compensation; 
        elsif repeat_pulse = '1' and nozeroperiod then
          state   <= COUNTING;
          --Store the period four cycles before 
          counter <= unsigned(pulse_period)-repeat_pulse_delay_compensation;
        else
          state 	<= WAIT_ST;
        end if;
      when COUNTING =>
        if counter = 0 then
          state 	<= CAPTURE_PERIOD;
		  elsif dio_pulse_immed_stb_d2 = '1' then
			 state <= WAIT_ST;
        else
          state   <= COUNTING;
          counter <= counter-1;
        end if;
      when CAPTURE_PERIOD =>
          state <= TRIGGER;
      when TRIGGER =>
        state 		<= WAIT_ST;
        if(nozeroperiod) then
          repeat_pulse <= '1';
        else
          repeat_pulse <= '0';
        end if;
      when others =>
        state 		<= WAIT_ST;
    end case;
  end if;
  end process;
  
  output_process : process(rst_n_i, state)
  begin
    if (rst_n_i = '0') then
      pulse_output_o <= '0';
    else
      case state is
        when WAIT_ST =>
          pulse_output_o <= '0';
        when COUNTING =>
          pulse_output_o <= '0';
        when TRIGGER =>
          pulse_output_o <= '1';
        when others =>
          pulse_output_o <= '0';
      end case;
    end if;
  end process;

end Behavioral;

