-------------------------------------------------------------------------------
-- Title      : DIO common package
-- Project    : FMC DIO 5CH TTa
-------------------------------------------------------------------------------
-- File       : dio_common_top.vhd
-- Author     : Miguel Jimenez-Lopez
-- Company    : Seven Solutions
-- Created    : 2019-03-27
-- Last update: 2019-12-04
-- Platform   : FPGA-generics
-- Standard   : VHDL
-------------------------------------------------------------------------------
-- Description:
-- DIO common package with NIC or Etherbone capabilities.
-------------------------------------------------------------------------------
-- Copyright (c) 2019 CERN (www.cern.ch)
--
-- This source file is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2.1 of the License, or (at your option) any
-- later version.
--
-- This source is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-- PURPOSE.  See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General
-- Public License along with this source; if not, download it
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author          Description
-- 2019-03-27  1.0      klyone          Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.wishbone_pkg.all;
use work.wr_board_pkg.all;
use work.wr_dio_pkg.all;

package dio_common_top_pkg is

  type t_dio_cfg_mode is (DIO_NIC, DIO_ETHERBONE);

  function f_pick_fabric_iface_for_dio(mode : t_dio_cfg_mode) return t_board_fabric_iface;
  function f_pick_sdb_layout_for_dio(mode   : t_dio_cfg_mode) return t_sdb_record_array;

  component dio_common_top
    generic (
      g_dio_mode    : t_dio_cfg_mode := DIO_NIC;
      g_dpram_initf : string         := "../../ip_cores/wr-cores/bin/wrpc/wrc_phy8.bram";
      -- Simulation-mode enable parameter. Set by default (synthesis) to 0, and
      -- changed to non-zero in the instantiation of the top level DUT in the testbench.
      -- Its purpose is to reduce some internal counters/timeouts to speed up simulations.
      g_simulation  : integer        := 0
      );
    port (
      ---------------------------------------------------------------------------
      -- Clocks/resets
      ---------------------------------------------------------------------------

      -- Local oscillators
      clk_20m_vcxo_i : in std_logic;    -- 20MHz VCXO clock

      clk_125m_pllref_p_i : in std_logic;  -- 125 MHz PLL reference
      clk_125m_pllref_n_i : in std_logic;

      clk_125m_gtp_n_i : in std_logic;  -- 125 MHz GTP reference
      clk_125m_gtp_p_i : in std_logic;

      ---------------------------------------------------------------------------
      -- GN4124 PCIe bridge signals
      ---------------------------------------------------------------------------
      -- From GN4124 Local bus
      gn_rst_n      : in    std_logic;  -- Reset from GN4124 (RSTOUT18_N)
      -- PCIe to Local [Inbound Data] - RX
      gn_p2l_clk_n  : in    std_logic;  -- Receiver Source Synchronous Clock-
      gn_p2l_clk_p  : in    std_logic;  -- Receiver Source Synchronous Clock+
      gn_p2l_rdy    : out   std_logic;  -- Rx Buffer Full Flag
      gn_p2l_dframe : in    std_logic;  -- Receive Frame
      gn_p2l_valid  : in    std_logic;  -- Receive Data Valid
      gn_p2l_data   : in    std_logic_vector(15 downto 0);  -- Parallel receive data
      -- Inbound Buffer Request/Status
      gn_p_wr_req   : in    std_logic_vector(1 downto 0);  -- PCIe Write Request
      gn_p_wr_rdy   : out   std_logic_vector(1 downto 0);  -- PCIe Write Ready
      gn_rx_error   : out   std_logic;  -- Receive Error
      -- Local to Parallel [Outbound Data] - TX
      gn_l2p_clkn   : out   std_logic;  -- Transmitter Source Synchronous Clock-
      gn_l2p_clkp   : out   std_logic;  -- Transmitter Source Synchronous Clock+
      gn_l2p_dframe : out   std_logic;  -- Transmit Data Frame
      gn_l2p_valid  : out   std_logic;  -- Transmit Data Valid
      gn_l2p_edb    : out   std_logic;  -- Packet termination and discard
      gn_l2p_data   : out   std_logic_vector(15 downto 0);  -- Parallel transmit data
      -- Outbound Buffer Status
      gn_l2p_rdy    : in    std_logic;  -- Tx Buffer Full Flag
      gn_l_wr_rdy   : in    std_logic_vector(1 downto 0);  -- Local-to-PCIe Write
      gn_p_rd_d_rdy : in    std_logic_vector(1 downto 0);  -- PCIe-to-Local Read Response Data Ready
      gn_tx_error   : in    std_logic;  -- Transmit Error
      gn_vc_rdy     : in    std_logic_vector(1 downto 0);  -- Channel ready
      -- General Purpose Interface
      gn_gpio       : inout std_logic_vector(1 downto 0);  -- gn_gpio[0] -> GN4124 GPIO8
                                        -- gn_gpio[1] -> GN4124 GPIO9
      ---------------------------------------------------------------------------
      -- SPI interface to DACs
      ---------------------------------------------------------------------------

      plldac_sclk_o   : out std_logic;
      plldac_din_o    : out std_logic;
      pll25dac_cs_n_o : out std_logic;  --cs1
      pll20dac_cs_n_o : out std_logic;  --cs2

      ---------------------------------------------------------------------------
      -- SFP I/O for transceiver
      ---------------------------------------------------------------------------

      sfp_txp_o         : out   std_logic;
      sfp_txn_o         : out   std_logic;
      sfp_rxp_i         : in    std_logic;
      sfp_rxn_i         : in    std_logic;
      sfp_mod_def0_i    : in    std_logic;  -- sfp detect
      sfp_mod_def1_b    : inout std_logic;  -- scl
      sfp_mod_def2_b    : inout std_logic;  -- sda
      sfp_rate_select_o : out   std_logic;
      sfp_tx_fault_i    : in    std_logic;
      sfp_tx_disable_o  : out   std_logic;
      sfp_los_i         : in    std_logic;

      ---------------------------------------------------------------------------
      -- Onewire interface
      ---------------------------------------------------------------------------

      onewire_b : inout std_logic;

      ---------------------------------------------------------------------------
      -- UART
      ---------------------------------------------------------------------------

      uart_rxd_i : in  std_logic;
      uart_txd_o : out std_logic;

      ---------------------------------------------------------------------------
      -- Flash memory SPI interface
      ---------------------------------------------------------------------------

      flash_sclk_o : out std_logic;
      flash_ncs_o  : out std_logic;
      flash_mosi_o : out std_logic;
      flash_miso_i : in  std_logic;

      ---------------------------------------------------------------------------
      -- Miscellanous SPEC pins
      ---------------------------------------------------------------------------
      -- Red LED next to the SFP: blinking indicates that packets are being
      -- transferred.
      led_act_o  : out std_logic;
      -- Green LED next to the SFP: indicates if the link is up.
      led_link_o : out std_logic;

      button1_i : in std_logic;

      ---------------------------------------------------------------------------
      -- Digital I/O FMC Pins
      -- used in this design to output WR-aligned 1-PPS (in Slave mode) and input
      -- 10MHz & 1-PPS from external reference (in GrandMaster mode).
      ---------------------------------------------------------------------------

      -- Clock input from LEMO 5 on the mezzanine front panel. Used as 10MHz
      -- external reference input.
      dio_clk_p_i : in std_logic;
      dio_clk_n_i : in std_logic;

      -- Differential inputs, dio_p_i(N) inputs the current state of I/O (N+1) on
      -- the mezzanine front panel.
      dio_n_i : in std_logic_vector(4 downto 0);
      dio_p_i : in std_logic_vector(4 downto 0);

      -- Differential outputs. When the I/O (N+1) is configured as output (i.e. when
      -- dio_oe_n_o(N) = 0), the value of dio_p_o(N) determines the logic state
      -- of I/O (N+1) on the front panel of the mezzanine
      dio_n_o : out std_logic_vector(4 downto 0);
      dio_p_o : out std_logic_vector(4 downto 0);

      -- Output enable. When dio_oe_n_o(N) is 0, connector (N+1) on the front
      -- panel is configured as an output.
      dio_oe_n_o : out std_logic_vector(4 downto 0);

      -- Termination enable. When dio_term_en_o(N) is 1, connector (N+1) on the front
      -- panel is 50-ohm terminated
      dio_term_en_o : out std_logic_vector(4 downto 0);

      -- Two LEDs on the mezzanine panel. Only Top one is currently used - to
      -- blink 1-PPS.
      dio_led_top_o : out std_logic;
      dio_led_bot_o : out std_logic;

      -- DIO 1-wire bus
      dio_onewire_b  : inout std_logic;
      dio_sdn_n_o    : out   std_logic;
      dio_sdn_ck_n_o : out   std_logic;
      -- I2C interface for accessing FMC EEPROM. Deprecated, was used in
      -- pre-v3.0 releases to store WRPC configuration. Now we use Flash for this.
      dio_scl_b : inout std_logic;
      dio_sda_b : inout std_logic;

      -- Leap second 
      leap_second_value_i       : in std_logic_vector(15 downto 0);
      leap_second_flag_59_i     : in std_logic;
      leap_second_flag_61_i     : in std_logic;
      leap_second_flag_valid_i  : in std_logic
      );
  end component;

  -- Number of masters on the wishbone crossbar
  constant c_NUM_WB_MASTERS : integer := 1;

  -- Number of slaves on the primary wishbone crossbar
  constant c_NUM_WB_SLAVES : integer := 3;

  -- Primary Wishbone master(s) offsets
  constant c_WB_MASTER_PCIE : integer := 0;

  -- Primary Wishbone slave(s) offsets
  constant c_WB_SLAVE_WRC     : integer := 0;
  constant c_WB_SLAVE_NIC_VIC : integer := 1;
  constant c_WB_SLAVE_DIO     : integer := 2;

  -- sdb header address on primary crossbar
  constant c_SDB_ADDRESS : t_wishbone_address := x"00070000";

-- f_xwb_bridge_manual_sdb(size, sdb_addr)
-- Note: sdb_addr is the sdb records address relative to the bridge base address
  constant c_wrc_bridge_sdb : t_sdb_bridge :=
    f_xwb_bridge_manual_sdb(x"0003ffff", x"00030000");
  constant c_nic_bridge_sdb : t_sdb_bridge :=
    f_xwb_bridge_manual_sdb(x"0001ffff", x"00011000");
  constant c_wr_dio_bridge_sdb : t_sdb_bridge :=
    f_xwb_bridge_product_manual_sdb(x"00000fff", x"00000800", c_xwr_dio_sdb);

-- Primary wishbone crossbar layout (NIC)
  constant c_NIC_WB_LAYOUT : t_sdb_record_array(c_NUM_WB_SLAVES - 1 downto 0) := (
    c_WB_SLAVE_WRC     => f_sdb_embed_bridge(c_wrc_bridge_sdb, x"00000000"),  -- WRPC
    c_WB_SLAVE_NIC_VIC => f_sdb_embed_bridge(c_nic_bridge_sdb, x"00040000"),  -- NIC
    c_WB_SLAVE_DIO     => f_sdb_embed_bridge(c_wr_dio_bridge_sdb, x"00060000")  -- DIO
    );

-- Primary wishbone crossbar layout (NIC)
  constant c_VIC_ONLY_WB_LAYOUT : t_sdb_record_array(c_NUM_WB_SLAVES - 1 downto 0) := (
    c_WB_SLAVE_WRC     => f_sdb_embed_bridge(c_wrc_bridge_sdb, x"00000000"),  -- WRPC
    c_WB_SLAVE_NIC_VIC => f_sdb_embed_device(c_xwb_vic_sdb, x"00040000"),  -- VIC
    c_WB_SLAVE_DIO     => f_sdb_embed_bridge(c_wr_dio_bridge_sdb, x"00060000")  -- DIO
    );

end dio_common_top_pkg;

package body dio_common_top_pkg is

  function f_pick_fabric_iface_for_dio(mode : t_dio_cfg_mode)
    return t_board_fabric_iface is
  begin
    case mode is
      when DIO_NIC       => return PLAIN;
      when DIO_ETHERBONE => return ETHERBONE;
      when others        => return PLAIN;  --by default
    end case;
  end function;

  function f_pick_sdb_layout_for_dio(mode : t_dio_cfg_mode)
    return t_sdb_record_array is
  begin
    case mode is
      when DIO_NIC       => return c_NIC_WB_LAYOUT;
      when DIO_ETHERBONE => return c_VIC_ONLY_WB_LAYOUT;
      when others        => return c_NIC_WB_LAYOUT;  --by default
    end case;
  end function;

end dio_common_top_pkg;
