/*
 * Copyright (C) 2020 CERN (www.cern.ch)
 *
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/timex.h>
#include <sys/ioctl.h>

#include "fmc-dio-device.h"
#include "fmc-dio-device-private.h"

fmc_dio_device create_fmc_dio_device(char *name)
{
	fmc_dio_device dev = calloc(1, sizeof(*dev));
	if(!dev)
		return BAD_FMC_DEVICE;

	strncpy(dev->dev_name, name, MAX_FMC_DIO_PATH-1);

	return dev;
}

static int check_fmc_dio_device(fmc_dio_device dev)
{
	if(dev != BAD_FMC_DEVICE) {
		return 0;
	}
	else {
		return 1;
	}
}

int open_fmc_dio_device(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return -EINVAL;

	dev->fd = open(dev->dev_name, O_RDWR);
	if(dev->fd < 0) {
		free(dev);
		return -ENODEV;
	}

	if(ioctl(dev->fd, PRIV_MEZZANINE_ID, (unsigned long) dev->dev_name) < 0 &&
		errno != EAGAIN) {
		free(dev);
		return -EFAULT;
	}

	return 0;
}

void attach_log_devices_to_fmc_dio_device(fmc_dio_device dev,
				log_device *log_devs,
				unsigned int n_log_devs)
{
	if(check_fmc_dio_device(dev) || !log_devs || n_log_devs == 0)
		return;

	dev->logs = log_devs;
	dev->n_logs = n_log_devs;
}

int enable_log_for_fmc_dio_device(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return -EINVAL;

	dev->enabled_log = 1;

	return 0;
}
int disable_log_for_fmc_dio_device(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return -EINVAL;

	dev->enabled_log = 0;

	return 0;
}

void close_fmc_dio_device(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return;

	close(dev->fd);
	free(dev);
}

int enable_fmc_dio_device_irq(fmc_dio_device dev, int ch)
{
	int ret;

	if(check_fmc_dio_device(dev)) {
		return -EINVAL;
	}

	ret = enable_fmc_dio_device_hw_irq(dev, ch, 1);

	return ret;
}

int enable_fmc_dio_device_all_irq(fmc_dio_device dev)
{
	int ret = 0;

	for(int i = 0 ; i <= FMC_DIO_N_CH ; i++) {
		ret = enable_fmc_dio_device_irq(dev, i);
		if(ret)
			break;
	}
	return ret;
}

int disable_fmc_dio_device_irq(fmc_dio_device dev, int ch)
{
	int ret;

	if(check_fmc_dio_device(dev)) {
		return -EINVAL;
	}

	ret = enable_fmc_dio_device_hw_irq(dev, ch, 0);

	return ret;
}

int disable_fmc_dio_device_all_irq(fmc_dio_device dev)
{
	int ret = 0;

	for(int i = 0 ; i <= FMC_DIO_N_CH ; i++) {
		ret = disable_fmc_dio_device_irq(dev, i);
		if(ret)
			break;
	}
	return ret;
}

static int enable_fmc_dio_device_hw_irq(fmc_dio_device dev,
					int ch, int en)
{
	if(ch < 0 || ch > FMC_DIO_N_CH) {
		return -EINVAL;
	}

	struct wr_dio_cmd *c = &dev->cmd;

	en = (en) ? 1 : 0;

	c->command = WR_DIO_CMD_MASK_IRQ;
	c->flags = WR_DIO_F_MASK;
	c->channel = 1 << ch;
	c->value = (en) ? WR_DIO_F_MASK_ENABLE_IRQ :
		WR_DIO_F_MASK_DISABLE_IRQ;

	if(ioctl(dev->fd, PRIV_MEZZANINE_CMD, c) < 0)
		return -EFAULT;
	return 0;
}

int setup_fmc_dio_device_irq(fmc_dio_device dev,
			unsigned long period,
			long count)
{
	int ret;

	if(check_fmc_dio_device(dev))
		return -EINVAL;

	ret = setup_fmc_dio_device_hw_irq(dev, period, count);

	return ret;
}

static int setup_fmc_dio_device_hw_irq(fmc_dio_device dev,
				unsigned long period,
				long count)
{
	struct wr_dio_cmd *c = &dev->cmd;
	int ret;

	c->command = WR_DIO_CMD_IRQ;
	c->flags = 0x0;
	c->flags |= WR_DIO_F_LOOP;
	c->flags |= WR_DIO_F_NOW;
	c->t[1].tv_nsec = period % NS_IN_A_SEC;
	c->t[1].tv_sec = period / NS_IN_A_SEC;
	c->value= count;

	ret = ioctl(dev->fd, PRIV_MEZZANINE_CMD, c);
	if(ret < 0) {
		return -EFAULT;
	}
	return 0;
}

int get_tai_ts_from_fmc_dio_device(fmc_dio_device dev, int ch,
				struct usr_timestamp **ts, unsigned int *nts)
{
	if(check_fmc_dio_device(dev))
		return -EINVAL;

	return get_hw_ts_from_fmc_dio_device(dev, ch, ts, nts);
}

int get_utc_ts_from_fmc_dio_device(fmc_dio_device dev, int ch,
				struct usr_timestamp **ts, unsigned int *nts)
{
	int ret;
	int kernel_tai;
	int usr_tai;
	int usr_tai_valid;
	int tai;
	struct timespec *t;

	ret = get_tai_ts_from_fmc_dio_device(dev, ch, ts, nts);
	if(ret)
		return ret;

	kernel_tai = get_kernel_leaps_info(dev);

	for(int i = 0 ; i < *nts ; i++) {
		t = &((*ts)[i].t);
		usr_tai = (*ts)[i].leap_second;
#if defined(IGNORE_USR_LEAPSECOND)
		usr_tai_valid = 0;
#else
		usr_tai_valid = (*ts)[i].leap_second_valid;
#endif
		tai = (usr_tai_valid) ? usr_tai : kernel_tai;
		t->tv_sec -= tai;
	}

	return ret;
}

static int get_hw_ts_from_fmc_dio_device(fmc_dio_device dev, int ch,
					struct usr_timestamp **ts, unsigned int *nts)
{
	struct wr_dio_cmd *c = &dev->cmd;
	int ret;

	c->command = WR_DIO_CMD_STAMP;
	c->flags = 0;
	c->channel = ch;

	ret = ioctl(dev->fd, PRIV_MEZZANINE_CMD, (unsigned long) c);
	if(ret < 0) {
		return -EFAULT;
	}

	if(c->nstamp <= 0) {
		return -EAGAIN;
	}

	*nts = c->nstamp;
	*ts = calloc(c->nstamp, sizeof(struct usr_timestamp));
	if(!*ts)
		return -ENOMEM;

	for(int i = 0 ; i < c->nstamp ; i++) {
		(*ts)[i].t = (c->t[i]);
		(*ts)[i].leap_second = (c->wr_ts_info_buf[i]).leap_second;
		(*ts)[i].leap_second_valid = (c->wr_ts_info_buf[i]).leap_second_valid;
		(*ts)[i].flag59 = (c->wr_ts_info_buf[i]).flag59;
		(*ts)[i].flag61 = (c->wr_ts_info_buf[i]).flag61;
	}

	dev->total_n_ts += c->nstamp;
	dev->partial_n_ts += c->nstamp;

	return 0;
}

static int get_kernel_leaps_info(fmc_dio_device dev)
{
	struct timex tx = {0};
	if (adjtimex(&tx) < 0) {
		return -EFAULT;
	}
	return tx.tai;
}

unsigned int get_total_n_timestamps(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return 0;

	return dev->total_n_ts;
}

unsigned int get_partial_n_timestamps(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return 0;

	return dev->partial_n_ts;
}

void reset_partial_n_timestamps(fmc_dio_device dev)
{
	if(check_fmc_dio_device(dev))
		return;

	dev->partial_n_ts = 0;
}

static void log_msg(fmc_dio_device dev, const char *msg)
{
	if(!dev->logs || dev->n_logs == 0)
		return;

	if(!dev->enabled_log)
		return;

	for(int i = 0 ; i < dev->n_logs ; i++)
		send_to_log_device(dev->logs[i], msg);
}

static void LOG(fmc_dio_device dev, const char *fmt, ...)
{
	va_list ap;
	char buffer[MAX_TMP_BUF] = {0};

	va_start(ap, fmt);
	vsnprintf(buffer, MAX_TMP_BUF-1, fmt, ap);
	va_end(ap);

	log_msg(dev, buffer);
}


