/*
 * Copyright (C) 2020 CERN (www.cern.ch)
 *
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>
#include <errno.h>

#include "stats-engine.h"
#include "stats-engine-private.h"

stats_engine create_stats_engine(void)
{
	stats_engine engine = calloc(1, sizeof(*engine));

	return engine;
}

int check_stats_engine(stats_engine engine)
{
	if(engine != BAD_STATS_ENGINE)
		return 0;
	else
		return 1;
}

void attach_log_devices_to_stats_engine(stats_engine engine,
				log_device *log_devs,
				unsigned int n_log_devs)
{
	if(check_stats_engine(engine) || !log_devs || n_log_devs == 0)
		return;

	engine->logs = log_devs;
	engine->n_logs = n_log_devs;
}


int enable_log_for_stats_engine(stats_engine engine)
{
	if(check_stats_engine(engine))
		return -EINVAL;

	engine->enabled_log = 1;

	return 0;
}

int disable_log_for_stats_engine(stats_engine engine)
{
	if(check_stats_engine(engine))
		return -EINVAL;

	return 0;
}

int add_usr_timestamp_to_stats_engine(stats_engine engine, struct usr_timestamp *ts)
{
	struct timespec system_ts;

	if(check_stats_engine(engine))
		return -EINVAL;

	clock_gettime(CLOCK_REALTIME, &system_ts);
	engine->sys_ts[engine->next_ts] = system_ts;
	engine->usr_ts[engine->next_ts] = *ts;

	update_index_stats_engine(&engine->next_ts);
	update_n_ts_stats_engine(engine);

	return 0;
}

int run_stats_engine(stats_engine engine, int verbose)
{
	struct timespec ts;
	int neg;
	int count_ts;

	if(check_stats_engine(engine))
		return -EINVAL;

	neg = timespec_subtract(&ts, &engine->sys_ts[engine->current_ts],
				&(engine->usr_ts[engine->current_ts].t));
	engine->diff_ts[engine->current_ts] = ((neg) ? -1.0 : 1.0)*(ts.tv_sec + ((double) ts.tv_nsec / NS_IN_A_SEC));
	engine->stats.mean = compute_mean(engine->diff_ts, engine->n_ts);
	engine->stats.std = compute_stdev(engine->diff_ts, engine->n_ts, engine->stats.mean);
	count_ts = compute_ts_per_second(engine->usr_ts[engine->current_ts].t.tv_sec,
							engine->usr_ts, engine->n_ts);
	if(count_ts != -1)
		engine->stats.ts_per_second = count_ts;

	engine->stats.ts_total++;
	log_stats_engine(engine, verbose);

	update_index_stats_engine(&engine->current_ts);

	engine->n_iter++;

	return 0;
}

void destroy_stats_engine(stats_engine engine)
{
	if(check_stats_engine(engine))
		return;

	free(engine);
}

static void update_n_ts_stats_engine(stats_engine engine)
{
	if(engine->n_ts < MAX_TS)
		engine->n_ts++;
}

static void update_index_stats_engine(unsigned int *index)
{
	if(*index == MAX_TS-1)
		*index = 0;
	else
		(*index)++;
}

static void log_stats_engine(stats_engine engine, int verbose)
{
	LOG(engine, "\n===================== STATS ENGINE LOG =====================\n");

	if(verbose) {
		LOG(engine, "Iterations: %d\n", engine->n_iter);
		LOG(engine, "Timestamps in buffer: %d", engine->n_ts);
		LOG(engine, "Next timestamp (index): %d\n", engine->next_ts);

		for(int i = 0 ; i < engine->n_ts ; i++) {
			LOG(engine, "\tsys_ts[%d]: sec %ld, nsec %ld",
				i, engine->sys_ts[i].tv_sec,
				engine->sys_ts[i].tv_nsec);
			LOG(engine, "\tusr_ts[%d]: sec %ld, nsec %ld [ls_valid: %d, ls: %d, l59: %d, l61: %d]",
				i, engine->usr_ts[i].t.tv_sec,
				engine->usr_ts[i].t.tv_nsec,
				engine->usr_ts[i].leap_second_valid,
				engine->usr_ts[i].leap_second,
				engine->usr_ts[i].flag59,
				engine->usr_ts[i].flag61);
			LOG(engine, "\tdiff[%d]: %f %s",
				i, engine->diff_ts[i], METRICS_UNIT);
		}
	}

	LOG(engine, "\nMetrics => Total timestamps: %d, Timestamps per second: %d\n"
		"mean: %f %s, stdev: %f %s",
		engine->stats.ts_total, engine->stats.ts_per_second, engine->stats.mean,
		METRICS_UNIT, engine->stats.std, METRICS_UNIT);
	LOG(engine, "\n============================================================\n");
	log_clear(engine);
}

static int check_log_configuration(stats_engine engine)
{
	if(!engine->logs || engine->n_logs == 0)
		return 1;

	if(!engine->enabled_log)
		return 1;

	return 0;
}

static void log_msg(stats_engine engine, const char *msg)
{
	if(check_log_configuration(engine))
		return;

	for(int i = 0 ; i < engine->n_logs ; i++)
		send_to_log_device(engine->logs[i], msg);
}

static void log_clear(stats_engine engine)
{
	if(check_log_configuration(engine))
		return;

	for(int i = 0 ; i < engine->n_logs ; i++)
		clear_log_device(engine->logs[i]);
}

static void LOG(stats_engine engine, const char *fmt, ...)
{
	va_list ap;
	char buffer[MAX_TMP_BUF] = {0};

	va_start(ap, fmt);
	vsnprintf(buffer, MAX_TMP_BUF-1, fmt, ap);
	va_end(ap);

	log_msg(engine, buffer);
}

static int timespec_subtract(struct timespec *result,
			struct timespec *x,
			struct timespec *y)
{
	struct timespec xx = *x;
	struct timespec yy = *y;
	x = &xx; y = &yy;

	if (x->tv_nsec > NS_IN_A_SEC-1)
	{
		x->tv_sec += x->tv_nsec / NS_IN_A_SEC;
		x->tv_nsec %= NS_IN_A_SEC;
	}

	if (y->tv_nsec > NS_IN_A_SEC-1)
	{
		y->tv_sec += y->tv_nsec / NS_IN_A_SEC;
		y->tv_nsec %= NS_IN_A_SEC;
	}

	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_nsec = x->tv_nsec - y->tv_nsec;

	if(result->tv_sec>0 && result->tv_nsec < 0)
	{
		result->tv_nsec += NS_IN_A_SEC;
		result->tv_sec--;
	}
	else if(result->tv_sec<0 && result->tv_nsec > 0)
	{
		result->tv_nsec -= NS_IN_A_SEC;
		result->tv_sec++;
	}

	return (result->tv_sec < 0) || (result->tv_nsec<0);
}

static double compute_mean(const double *values,
			const unsigned int n_values)
{
	double mean = 0.0;

	for(int i = 0 ; i < n_values ; i++) {
		mean += values[i];
	}

	mean /= n_values;

	return mean;
}

static double compute_stdev(const double *values,
			const unsigned int n_values,
			const double mean)
{
	double stdev = 0.0;
	double aux = 0.0;

	for(int i = 0 ; i < n_values ; i++) {
		aux += pow((values[i]-mean),2);
	}

	stdev = sqrt((aux/n_values));

	return stdev;
}

static int compute_ts_per_second(const int sec,
				const struct usr_timestamp *ts,
				const unsigned int n_ts)
{
	static int prev_sec = -1;
	int count = -1;

	if(prev_sec == -1) {
		prev_sec = sec;
	}
	else {
		if(prev_sec != sec) {
			count = 0;
			for(int i = 0 ; i < n_ts ; i++) {
				if(ts[i].t.tv_sec == prev_sec)
					count++;
			}
			prev_sec = sec;
		}
	}

	return count;
}
