/*
 * Copyright (C) 2019 CERN (www.cern.ch)
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <linux/module.h>
#include <linux/version.h>
#if KERNEL_VERSION(4,11,0) > LINUX_VERSION_CODE
#include <linux/sched.h>
#else
#include <linux/sched/signal.h>
#endif
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/ktime.h>
#include <linux/atomic.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/fmc.h>
#include <linux/fmc-sdb.h>

#include <linux/delay.h>

#include "fmc-dio.h"

#ifdef DIO_STAT
#define fmc_dio_stat 1
#else
#define fmc_dio_stat 0
#endif

#define CHANNEL_5_IRQ_EN_MASK 0x100000

#define FMC_DIO_IRQ_MASK  \
	(DIO_EIC_ISR_NEMPTY_0 \
	| DIO_EIC_ISR_NEMPTY_1\
	| DIO_EIC_ISR_NEMPTY_2\
	| DIO_EIC_ISR_NEMPTY_3\
	| DIO_EIC_ISR_NEMPTY_4)

#define FMC_DIO_IRQ_MASK_V2  \
	(DIO_EIC_ISR_NEMPTY_0 \
	| DIO_EIC_ISR_NEMPTY_1\
	| DIO_EIC_ISR_NEMPTY_2\
	| DIO_EIC_ISR_NEMPTY_3\
	| DIO_EIC_ISR_NEMPTY_4\
	| DIO_EIC_ISR_NEMPTY_5)

/* This is the structure we need to manage interrupts and loop internally */
#define FMC_DIO_BUFFER_LEN  512
struct dio_channel {
	struct timespec tsbuf[FMC_DIO_BUFFER_LEN];
	struct wr_timestamp_info wr_ts_info_buf[FMC_DIO_BUFFER_LEN];
	int bhead, btail;
	wait_queue_head_t q;

	/* The input event may fire a new pulse on this or another channel */
	struct timespec prevts, delay;
	atomic_t count;
	int target_channel;
};

struct dio_device {
	struct dio_channel ch[6];
};

/* Instead of timespec_sub, just subtract the nanos */
static inline void fmc_dio_int_ts_sub(struct timespec *ts, int nano)
{
	set_normalized_timespec(ts, ts->tv_sec, ts->tv_nsec - nano);
}

/* This programs a new pulse without changing the width */
static void __fmc_dio_int_new_pulse(struct fmc_dio *dev, int ch,
			    struct timespec *ts)
{
	void __iomem *base = dev->dio;
	struct regmap *map;

	struct regmap_common dio;
	dio = get_regmap_common(dev->version);

	map = get_regmap(dev->version);
	map += ch;

	fmc_dio_int_ts_sub(ts, 16); /* 2 cycles, to account for output latencies */
	writel(ts->tv_nsec / 8, base + map->cycle);
	writel(GET_HI32(ts->tv_sec), base + map->trig_h);
	writel(ts->tv_sec, base + map->trig_l);

	writel(1 << ch, base + dio.latch_reg);
}

static int fmc_dio_int_cmd_pulse(struct fmc_dio *dev,
			   struct wr_dio_cmd *cmd)
{
	void __iomem *base = dev->dio;
	struct PPSG_WB __iomem *ppsg = dev->ppsg;
	struct dio_device *d = dev->priv;
	struct dio_channel *c;
	struct regmap *map;
	struct timespec *ts;
	uint32_t reg;
	int ch;
	struct regmap_common dio;
	uint32_t pulse_per_loaded;

	dio = get_regmap_common(dev->version);

	ch = cmd->channel;
	if (ch > 4)
		return -EINVAL; /* mask not supported */
	c = d->ch + ch;

	map = get_regmap(dev->version);
	map += ch;

	ts = cmd->t;

	/* First, configure this bit as DIO output */
	reg = readl(base + dio.iomode_reg);
	writel(reg | (1 << 4*ch), base + dio.iomode_reg);

	/* Clear width and period registers */
	pulse_per_loaded = readl(base + map->pulse_per);
	writel(0, base + map->pulse_per);
	writel(1, base + map->pulse); //Set to 1 to avoid 3 second pulse due to compensation into gw
	writel(1 << ch, base + dio.latch_reg);
	atomic_set(&c->count, 0);

	if(cmd->flags & WR_DIO_F_REL)
	{
		pulse_per_loaded *= 8;
		msleep(pulse_per_loaded/1000000);
	}


	if(cmd->value == 0 && cmd->flags & WR_DIO_F_LOOP)
		return 0;

	writel(ts[1].tv_nsec / 8, base + map->pulse); /* width */

	//if(cmd->value == 0)
	//	return 0;


	if (cmd->flags & WR_DIO_F_NOW) {
		/* Generate a pulse train from current time in V2 DIO version*/
		if(dev->version == 1) {
			if (cmd->flags & WR_DIO_F_LOOP && cmd->value != 1) {
				writel(ts[2].tv_nsec / 8, base + map->pulse_per);
				c->target_channel = ch;

				cmd->value-=2;
				
				atomic_set(&c->count, cmd->value);
			}
		}
		/* if "now" we are done */
		writel(1 << ch, base + dio.pulse_reg);		
		return 0;
	}
	/* if relative, add current 40-bit second to timespec */
	if (cmd->flags & WR_DIO_F_REL) {  	
		uint32_t h1, l, h2;
		unsigned long now;
		h1 = readl(&ppsg->CNTR_UTCHI);
		l = readl(&ppsg->CNTR_UTCLO);
		h2 = readl(&ppsg->CNTR_UTCHI);
		if (h2 != h1)
			l = readl(&ppsg->CNTR_UTCLO);
		now = l;
		SET_HI32(now, h2);
		ts->tv_sec += now;
	}

	if (cmd->flags & WR_DIO_F_LOOP && cmd->value != 1) {
		c->target_channel = ch;
			
		if(dev->version == 0) {
			c->delay = ts[2];
			/* c->count is used after the pulse, so remove the first */
			if (cmd->value > 0)
				cmd->value--;
		}
		else {
			writel(ts[2].tv_nsec / 8, base + map->pulse_per);
			cmd->value-=2;
		}

		atomic_set(&c->count, cmd->value);
		c->prevts = ts[0]; /* our current setpoint */
	}

	__fmc_dio_int_new_pulse(dev, ch, ts);
	return 0;
}

static int fmc_dio_int_cmd_update_width(struct fmc_dio *dev,
			   struct wr_dio_cmd *cmd)
{
	void __iomem *base = dev->dio;
	struct dio_device *d = dev->priv;
	struct dio_channel *c;
	struct regmap *map;
	struct timespec *ts;
	int ch;
	struct regmap_common dio;
	uint32_t period;

	if(dev->version == 0){
		return -EINVAL;
	}

	dio = get_regmap_common(dev->version);

	ch = cmd->channel;
	if (ch > 4)
		return -EINVAL; /* mask not supported */
	c = d->ch + ch;

	map = get_regmap(dev->version);
	map += ch;

	ts = cmd->t;

	period = readl(base + map->pulse_per);
	if(ts[1].tv_nsec / 8 < period)
	{
		writel(ts[1].tv_nsec / 8, base + map->pulse);
		writel(1 << ch, base + dio.latch_reg);
	}
	else
		return -EINVAL;

	return 0;
}

static int fmc_dio_int_cmd_stamp(struct fmc_dio *dev,
			     struct wr_dio_cmd *cmd)
{
	struct dio_device *d = dev->priv;
	struct dio_channel *c = 0;
	struct timespec *ts = cmd->t;
	struct wr_timestamp_info *ts_info = cmd->wr_ts_info_buf;
	struct regmap *map;
	int mask, ch, last;
	int nstamp = 0;

	if ((cmd->flags & (WR_DIO_F_MASK || WR_DIO_F_WAIT))
	    == (WR_DIO_F_MASK || WR_DIO_F_WAIT))
		return -EINVAL; /* wait on several channels not supported */

again:
	if (cmd->flags & WR_DIO_F_MASK) {
		ch = 0;
		if(dev->version == 0) 
			last = 4;
		else
			last = 5;
		mask = cmd->channel;
	} else {
		ch = cmd->channel;
		last = ch;
		mask = (1 << ch);
	}
	/* handle the 1-channel and mask case in the same loop */
	c = d->ch + ch;
	for (; ch <= last; ch++, c++) {
		if (((1 << ch) & mask) == 0)
			continue;

		map = get_regmap(dev->version);
		map += ch;

		while (1) {
			if (nstamp == WR_DIO_N_STAMP)
				break;
			if (c->bhead == c->btail)
				break;
			*ts = c->tsbuf[c->btail];
			*ts_info = c->wr_ts_info_buf[c->btail];
			c->btail = (c->btail + 1) % FMC_DIO_BUFFER_LEN;
			nstamp++;
			ts++;
			ts_info++;
		}
		if (nstamp) {
			cmd->channel = ch;
			break;
		}
	}
	cmd->nstamp = nstamp;

	/* The user may asketo wait for timestamps, but for 1 channel only */
	if (!nstamp && cmd->flags & WR_DIO_F_WAIT) {
		ch--; c--; /* The for above incremeted them */
		try_module_get(THIS_MODULE);
		wait_event_interruptible(c->q, c->bhead != c->btail);
		module_put(THIS_MODULE);
		if (signal_pending(current))
			return -ERESTARTSYS;
		goto again;
	}

	if (!nstamp)
		return -EAGAIN;
	return 0;
}

static int fmc_dio_int_cmd_inout(struct fmc_dio *dev,
			     struct wr_dio_cmd *cmd)
{
	struct fmc_dio_gpio_block __iomem *gpio = dev->gpio;
	int mask, ch, last, bits;
	uint32_t reg, iomode;
	void __iomem *base = dev->dio;

	struct regmap_common dio;
	dio = get_regmap_common(dev->version);

	if (cmd->flags & WR_DIO_F_MASK) {
		ch = 0;
		last = 4;
		mask = cmd->channel;
	} else {
		ch = cmd->channel;
		last = ch;
		mask = (1 << ch);
		cmd->value <<= ch;
	}

	/* handle the 1-channel and mask case in the same loop */
	for (; ch <= last; ch++) {
		if (((1 << ch) & mask) == 0)
			continue;
		/* select the bits by shifting back the value field */
		bits = cmd->value >> ch;

		/* Obtain the current value in iomode */
		reg = readl(base + dio.iomode_reg) & ~(0xF << 4*ch);

		/* Select IO mode */
		if (bits & WR_DIO_INOUT_DIO) {
			if(bits & WR_DIO_INOUT_VALUE)
				iomode = 2; /* WRPC connection */
			else
				iomode = 1; /* DIO connection */
		} else {
			iomode = 0; /* GPIO  connection */

			/* Output value is bit 0 (0x1) */
			if (bits & WR_DIO_INOUT_VALUE)
				writel(FMC_DIO_GPIO_VALUE(ch), &gpio->set);
			else
				writel(FMC_DIO_GPIO_VALUE(ch), &gpio->clear);
		}

		/* Appends to iomode TERM and OUTPUT_ENABLE_N bits */
		iomode |= (((bits & WR_DIO_INOUT_TERM) != 0) << 3)
			| (((bits & WR_DIO_INOUT_OUTPUT) == 0) << 2);
		writel(reg | (iomode << 4*ch), base + dio.iomode_reg);
	}
	return 0;
}

static int fmc_dio_int_cmd_mask_irq(struct fmc_dio *dev,
			     struct wr_dio_cmd *cmd)
{
	struct fmc_dio_gpio_block __iomem *gpio = dev->gpio;
	int mask, ch, last, bits;
	uint32_t reg, iomode;
	void __iomem *base = dev->dio;

	if(dev->version == 0){ //Not implemented in old version
		return -EINVAL;
	}

	struct regmap_common dio;
	dio = get_regmap_common(dev->version);

	if (cmd->value == WR_DIO_F_MASK_READ_IRQ) {
		reg = readl(base + dio.eic_imr_reg);
		if(reg & cmd->channel) {
			cmd->value = WR_DIO_F_MASK_ENABLE_IRQ;
			return 0;
		} else {
			cmd->value = WR_DIO_F_MASK_DISABLE_IRQ;
			return 0;
		}
	} else if (cmd->value == WR_DIO_F_MASK_ENABLE_IRQ) {
		writel(cmd->channel, base + dio.eic_ier_reg);
	} else if (cmd->value == WR_DIO_F_MASK_DISABLE_IRQ) {
		writel(cmd->channel, base + dio.eic_idr_reg);
	} else {
		return -1;
	}
	
	return 0;
}

static int fmc_dio_int_cmd_irq(struct fmc_dio *dev,
			     struct wr_dio_cmd *cmd)
{

	void __iomem *base = dev->dio;
	struct PPSG_WB __iomem *ppsg = dev->ppsg;
	struct dio_device *d = dev->priv;
	struct dio_channel *c;
	struct regmap *map;
	struct timespec *ts;
	uint32_t reg;
	int ch;
	struct regmap_common dio;
	uint32_t pulse_per_loaded;

	if(dev->version != 1)
		return -ENOTSUPP;
	

	
	dio = get_regmap_common(dev->version);

	ch = 5;

	c = d->ch + ch;

	map = get_regmap(dev->version);
	map += ch;

	ts = cmd->t;

	pulse_per_loaded = readl(base + map->pulse_per);
	writel(0, base + map->pulse_per);
	writel(1, base + map->pulse); //Set to 1 to avoid 3 second pulse due to compensation into gw
	writel(1 << ch, base + dio.latch_reg);
	atomic_set(&c->count, 0);

	if(cmd->flags & WR_DIO_F_REL)
	{
		pulse_per_loaded *= 8;
		msleep(pulse_per_loaded/1000000);
	}

	if(cmd->value == 0 && cmd->flags & WR_DIO_F_LOOP)
		return 0;

	/* First, configure the IRQ channel */
	reg = readl(base + dio.iomode_reg);
	writel((reg | CHANNEL_5_IRQ_EN_MASK), base + dio.iomode_reg);
	writel(100, base + map->pulse); /* width */

	if (cmd->flags & WR_DIO_F_NOW) {
		/* Generate a pulse train from current time*/	
		if (cmd->flags & WR_DIO_F_LOOP && cmd->value != 1) {
			writel(ts[1].tv_nsec / 8, base + map->pulse_per);
			c->target_channel = ch;

			cmd->value-=2;
			
			atomic_set(&c->count, cmd->value);
		}
		writel(1 << ch, base + dio.pulse_reg);		
		return 0;
	}
	/* if relative, add current 40-bit second to timespec */
	if (cmd->flags & WR_DIO_F_REL) {  	
		uint32_t h1, l, h2;
		unsigned long now;
		h1 = readl(&ppsg->CNTR_UTCHI);
		l = readl(&ppsg->CNTR_UTCLO);
		h2 = readl(&ppsg->CNTR_UTCHI);
		if (h2 != h1)
			l = readl(&ppsg->CNTR_UTCLO);
		now = l;
		SET_HI32(now, h2);
		ts->tv_sec += now;
	}

	if (cmd->flags & WR_DIO_F_LOOP && cmd->value != 1) {
		c->target_channel = ch;

		/* Configure period */	
		writel( (ts[1].tv_nsec / 8) + (ts[1].tv_sec * 1000000000 / 8), base + map->pulse_per);
		cmd->value-=2;

		atomic_set(&c->count, cmd->value);
		c->prevts = ts[0]; /* our current setpoint */
	}

	__fmc_dio_int_new_pulse(dev, ch, ts);

	return 0;

}


int fmc_dio_int_ioctl(struct fmc_dio *dev, unsigned int ioctlcmd,
		unsigned long arg)
{
	struct wr_dio_cmd *cmd;
	struct fmc_device *fmc = dev->fmc;
	ktime_t t, t0;
	int ret;

	if (ioctlcmd == PRIV_MEZZANINE_ID)
		return -EAGAIN; /* Special marker */
	if (ioctlcmd != PRIV_MEZZANINE_CMD)
		return -ENOIOCTLCMD;

	if (fmc_dio_stat) {
		t0 = ktime_get();
	}

	/* The cmd struct can't fit in the stack, so allocate it */
	cmd = kmalloc(sizeof(*cmd), GFP_KERNEL);
	if (!cmd)
		return -ENOMEM;
	ret = -EFAULT;
	if (copy_from_user(cmd, (struct wr_dio_cmd *) arg, sizeof(*cmd)))
		goto out;

	switch(cmd->command) {
	case WR_DIO_CMD_PULSE:
		ret = fmc_dio_int_cmd_pulse(dev, cmd);
		break;
	case WR_DIO_CMD_STAMP:
		ret = fmc_dio_int_cmd_stamp(dev, cmd);
		break;
	case WR_DIO_CMD_INOUT:
		ret = fmc_dio_int_cmd_inout(dev, cmd);
		break;
	case WR_DIO_CMD_DAC:
		ret = -ENOTSUPP;
		goto out;
	case WR_DIO_CMD_IRQ:
		ret = fmc_dio_int_cmd_irq(dev, cmd);
		break;
	case WR_DIO_CMD_UPDATE_WIDTH:
		ret = fmc_dio_int_cmd_update_width(dev, cmd);
		break;
	case WR_DIO_CMD_MASK_IRQ:
		ret = fmc_dio_int_cmd_mask_irq(dev, cmd);
		break;
	default:
		ret = -EINVAL;
		goto out;
	}

	if (copy_to_user((struct wr_dio_cmd *)arg, cmd, sizeof(*cmd)))
		return -EFAULT;
out:
	kfree(cmd);

	if (fmc_dio_stat) {
		t = ktime_sub(ktime_get(), t0);
		dev_info(fmc->hwdev, "ioctl: %li ns\n", (long)ktime_to_ns(t));
	}
	return ret;
}

/* This is called from the interrupt handler to program a new pulse */
static void fmc_dio_int_trig_next_pulse(struct fmc_dio *dev, int ch,
				struct dio_channel *c, struct timespec *ts)
{
	struct timespec newts;

	if (c->target_channel == ch) {
		c->prevts = timespec_add(c->prevts, c->delay);
		newts = c->prevts;
	} else {
		newts = timespec_add(*ts, c->delay);
	}
	__fmc_dio_int_new_pulse(dev, c->target_channel, &newts);

	/* If the count is not-infinite, decrement it */
	if (atomic_read(&c->count) > 0)
		atomic_dec(&c->count);
}

irqreturn_t fmc_dio_int_interrupt(struct fmc_dio *dev)
{
	struct fmc_device *fmc = dev->fmc;
	void __iomem *base = dev->dio;
	struct dio_device *d = dev->priv;
	static ktime_t t_ini, t_end;
	static int rate_avg;
	struct dio_channel *c;
	struct timespec *ts;
	struct wr_timestamp_info *ts_info;
	struct regmap *map;
	uint32_t mask, reg;
	int ch, chm;

	struct regmap_common dio;
	dio = get_regmap_common(dev->version);

	if (unlikely(!fmc->eeprom)) {
		dev_err(fmc->hwdev, "WR-DIO: No mezzanine, disabling irqs\n");
		writel(~0, base + dio.eic_idr_reg);
		writel(~0, base + dio.eic_isr_reg);
		return IRQ_NONE;
	}

	/* Protect against interrupts taking 100% of cpu time */
	if (ktime_to_ns(t_end)) {
		int rate;
		u64 offtime, ontime;

		ontime = ktime_to_ns(t_end) - ktime_to_ns(t_ini);
		t_ini = ktime_get();
		offtime = ktime_to_ns(t_ini) - ktime_to_ns(t_end);

		/* avoid __udivdi3 */
		if (offtime > 100 * ontime)
			rate = 0;
		else
			rate = ((int)ontime * 100) / (int)offtime;

		rate_avg = (rate_avg * 1023 + rate) / 1024;
		if (rate_avg > 80) {
			dev_warn(fmc->hwdev, "DIO irq takes > 80%% CPU time: "
				 "disabling\n");
			disable_irq(dev->irq);
		}
	}

	if(dev->version == 0)
		mask = readl(base + dio.eic_isr_reg) & FMC_DIO_IRQ_MASK;
	else
		mask = readl(base + dio.eic_isr_reg) & FMC_DIO_IRQ_MASK_V2;

	/* Three indexes: channel, channel-mask, channel pointer */
	for (ch = 0, chm = 1, c = d->ch; mask; ch++, chm <<= 1, c++) {
		int h;

		if (!(mask & chm))
			continue;
		mask &= ~chm;

		/* Pull the FIFOs to the device structure */
		map = get_regmap(dev->version);
		map += ch;
		
		ts = NULL;
		ts_info = NULL;
		while (1) {
			reg = readl(base + map->fifo_status);
			if (reg & 0x20000) /* empty */
				break;
			h = c->bhead;
			ts = c->tsbuf + h;
			ts_info = c->wr_ts_info_buf + h;
			c->bhead = (h + 1) % FMC_DIO_BUFFER_LEN;
			if (c->bhead == c->btail)
				c->btail = (c->btail + 1) % FMC_DIO_BUFFER_LEN;
			/*
			 * fifo is not-empty, pick one sample. Read
			 * cycles last, as that operation pops the FIFO
			 */
			ts->tv_sec = 0;
			SET_HI32(ts->tv_sec, readl(base + map->fifo_tai_h));
			ts->tv_sec |= readl(base + map->fifo_tai_l);
			ts->tv_nsec = 8 * readl(base + map->fifo_cycle);

			uint32_t leap_second_reg = readl(base + map->fifo_leap_second);
			ts_info->leap_second = leap_second_reg & DIO_TSF_R3_LEAP_SECOND_VALUE_MASK;
			ts_info->leap_second_valid = (leap_second_reg >> DIO_TSF_R3_LEAP_SECOND_VALID_SHIFT) & 1;
			ts_info->flag59 = (leap_second_reg >> DIO_TSF_R3_LEAP_SECOND_FLAG_59_SHIFT) & 1;
			ts_info->flag61 = (leap_second_reg >> DIO_TSF_R3_LEAP_SECOND_FLAG_61_SHIFT) & 1;
			/* subtract 4 cycles lost in input sync circuits */
			fmc_dio_int_ts_sub(ts, 32);
		}
		writel(chm, base + dio.eic_isr_reg); /* ack */


		if(dev->version == 0) {
			if (ts && atomic_read(&c->count) != 0 ) {
				/* Only program new pulse if we are on the DIO V1 */
				fmc_dio_int_trig_next_pulse(dev, ch, c, ts);
			}
		}
		else 
		{
			if(atomic_read(&c->count) == 0) {
				/* If we are on the DIO V2, clear pulse period register 
				* when all the pulses have been generated
				*/
				writel(0, base + map->pulse_per);
				writel(1 << ch, base + dio.latch_reg);
			}
			else {
				/* If the count is not-infinite, decrement it */
				if (atomic_read(&c->count) > 0)
					atomic_dec(&c->count);
			}
		}

		wake_up_interruptible(&c->q);
	}
	t_end = ktime_get();
	return IRQ_HANDLED;
}

/* Init and exit below are called when a netdevice is created/destroyed */
int fmc_dio_internal_create(struct fmc_dio *dev)
{
	struct dio_device *d;
	void __iomem *base = dev->dio;
	int i;

	struct regmap_common dio;
	dio = get_regmap_common(dev->version);

	/* Allocate the data structure and enable interrupts for stamping */
	d = kzalloc(sizeof(*d), GFP_KERNEL);
	if (!d)
		return -ENOMEM;
	for (i = 0; i < ARRAY_SIZE(d->ch); i++)
		init_waitqueue_head(&d->ch[i].q);
	dev->priv = d;

	/*
	 * Enable interrupts for FIFO, if there's no mezzanine the
	 * handler will notice and disable the interrupts
	 */
	if(dev->version == 0)
		writel(FMC_DIO_IRQ_MASK, base + dio.eic_ier_reg);
	else
		writel(FMC_DIO_IRQ_MASK_V2, base + dio.eic_ier_reg);

	return 0;
}

void fmc_dio_internal_destroy(struct fmc_dio *dev)
{
	struct regmap_common dio;
	void __iomem *base = dev->dio;
	dio = get_regmap_common(dev->version);

	writel(~0, base + dio.eic_idr_reg);
	if (dev->priv)
		kfree(dev->priv);
}


