/*
 * Copyright (C) 2019 CERN (www.cern.ch)
 *
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 * Author: Alessandro Rubini <rubini@gnudd.com>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/firmware.h>
#include <linux/delay.h>
#include <linux/fmc.h>
#include <linux/fmc-sdb.h>

#include "fmc-dio.h"

/* Temporary pointer to the last FMC DIO */
static struct fmc_dio *last_dio;

/**
 * FMC DIO Interrupt Service Routing (ISR)
 *
 * @param irq IRQ number
 * @param dev_id Device identifier for the ISR
 *
 * @return IRQ_HANDLED if interrupt was handled properly and IRQ_NONE otherwise
 */
irqreturn_t fmc_dio_interrupt(int irq, void *dev_id)
{
	irqreturn_t ret;
	
	ret = fmc_dio_int_interrupt(dev_id);
	
	return ret;
}

// Macros for the memory-mapped hardware resources
#define FMC_DIO_RESOURCE_DIO 0
#define FMC_DIO_RESOURCE_GPIO 1
#define FMC_DIO_RESOURCE_PPSG 2
/**
 * This function creates the hardware resources for the FMC DIO.
 *
 * @param dio FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_resource_map(struct fmc_dio *dio)
{
	int i,err;
	struct platform_device *dev = dio->pdev;
	//struct fmc_device *fmc = dio->fmc;
	struct resource *r;
	void __iomem *mem;

	// Map the resources
	for(i = 0 ; i < dev->num_resources ; i++) {
		r = platform_get_resource(dev, IORESOURCE_MEM, i);
		if (!r || !r->start)
			continue;
		
		mem = ioremap(r->start, r->end + 1 - r->start);
		if (!mem) {
			dev_err(&dev->dev, "Remap for res %i %pr failed\n",
				i, r);
			return -ENOMEM;
		}

		switch(i)
		{
		case FMC_DIO_RESOURCE_DIO:
			dio->dio = mem;
			break;
		case FMC_DIO_RESOURCE_GPIO:
			dio->gpio = mem;
			break;
		case FMC_DIO_RESOURCE_PPSG:
			dio->ppsg = mem;
			break;
		default:
			/* Ugh? What is this? */
			break;
		};
	}

	// If some hardware resource is not detected, fail
	if(!dio->dio || !dio->gpio || !dio->ppsg) {
		printk("%s: ERROR in FPGA base pointers initialization. \n",__func__);
		return -EINVAL;
	}

	// Map the IRQ
	dio->irq = platform_get_irq(dev,0);
	if(dio->irq < 0) {
		printk("%s: ERROR in FMC DIO IRQ extraction. \n",__func__);
		return -EINVAL;
	}
	
	err = request_any_context_irq(dio->irq, fmc_dio_interrupt, 
		IRQF_TRIGGER_HIGH, dio->name, dio);
	if(err < 0) {
		printk("%s: ERROR in request IRQ (%d). \n",__func__,err);
		return err;
	}

	return 0;
}

/**
 * This function releases the hardware resources for the FMC DIO.
 *
 * @param dio FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_resource_release(struct fmc_dio *dio)
{
	// Release IRQ for FMC DIO
	free_irq(dio->irq, dio);

	return 0;
}

/**
 * FMC DIO probe function for the platform_device
 *
 * @param dev FMC DIO platform device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_pl_probe(struct platform_device *dev)
{
	/* NOTE: fmc_device is created by the same driver than platform_device
	 * and it had filled the private data structure pointer of the
	 * platform_device before registering it.
	 */
	struct fmc_device *fmc;
	struct fmc_dio *dio;
	int ret = 0;
	unsigned int devfn;
	unsigned char busn;
	
	dio = last_dio;
	if(!dio) {
		printk("%s: DIO internal structure is NULL \n",__func__);
		ret = -EINVAL;
		goto exit;
	}
	
	fmc = dio->fmc;	
	if(!fmc) {
		printk("%s: FMC device is NULL \n",__func__);
		ret = -EINVAL;
		goto exit;
	}
	
	/* Set up the FMC DIO name for the Linux Miscdevice */
	busn = (unsigned char) (fmc->device_id >> 8);
	devfn = (unsigned int) (fmc->device_id & 0xFF);
	snprintf(dio->name,FMC_DIO_NAME_MAX,
		FMC_DIO_NAME_PATTERN,
		busn,devfn);
	
	// Specific board info
	dio->board = dev->id_entry->driver_data;
	dio->pdev = dev;

	dio->version = dio->board;

	if(dio->version == FMC_DIO_BOARD_SPEC) {
		printk("DIO driver. Detected DIO V1");
	}	
	else if(dio->version == FMC_DIOv2_BOARD_SPEC) {
		printk("DIO driver. Detected DIO V2");
	}
	else{
		printk("DIO driver. DIO not detected");
	}

	/* Map the resources for the FMC DIO */
	ret = fmc_dio_resource_map(dio);
	if(ret)
		goto exit;
		
	ret = fmc_dio_internal_create(dio);
	if(ret)
		goto exit;

	/* Create a Linux Miscdevice for the FMC DIO */
	ret = fmc_dio_mdev_create(dio);
	if(ret)
		goto exit;

	/* Create Linux GPIO for the FMC DIO */
	ret = fmc_dio_gpio_init(dio);
	if(ret)
		goto exit;
	
exit:
	return ret;
}

/**
 * FMC DIO remove function for the platform_device
 *
 * @param dev FMC DIO platform device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_pl_remove(struct platform_device *dev)
{
	struct fmc_device *fmc;
	struct fmc_dio *dio;
	int ret = 0;
	
	dio = last_dio;
	fmc = dio->fmc;

	/* Release Linux GPIO for the FMC DIO */
	fmc_dio_gpio_exit(dio);

	/* Destroy the Linux Miscdevice for the FMC DIO */
	fmc_dio_mdev_destroy(dio);
	
	/* Destroy internal DIO structure */
	fmc_dio_internal_destroy(dio);

	/* Release the resources for the FMC DIO */
	ret = fmc_dio_resource_release(dio);
	if(ret)
		return ret;

	return ret;
}

/* Forward declarations */
int fmc_dio_probe(struct fmc_device *fmc);
int fmc_dio_remove(struct fmc_device *fmc);

/**
 * List of supported platform
 */
static const struct platform_device_id fmc_dio_id_table[] = {
	{	/* SPEC compatible */
		.name = "fmc-dio-spec",
		.driver_data = FMC_DIO_BOARD_SPEC,
	},
	{	/* SPEC compatible */
		.name = "fmc-diov2-spec",
		.driver_data = FMC_DIOv2_BOARD_SPEC,
	},
	{},
};

// FMC DIO platform driver
static struct platform_driver fmc_dio_pl_drv = {
	.probe		= fmc_dio_pl_probe,
	.remove		= fmc_dio_pl_remove,
	.id_table   = fmc_dio_id_table,
	/* No suspend or resume by now */
	.driver		= {
		.name		= "fmc-dio",
		.owner		= THIS_MODULE,
	},
};

// FMC DIO FRU identifier
static struct fmc_fru_id fd_fru_id[] = {
	{
		.product_name = "FmcDio5cha",
	},
};

// FMC DIO driver structure (FMC)
static struct fmc_driver fmc_dio_drv = {
	.version = FMC_VERSION,
	.driver.name = KBUILD_MODNAME,
	.probe = fmc_dio_probe,
	.remove = fmc_dio_remove,
	.id_table = {
		.fru_id = fd_fru_id,
		.fru_id_nr = ARRAY_SIZE(fd_fru_id),
	},
};

/**
 * FMC DIO probe function
 *
 * This function initializes the FMC DIO data structures.
 *
 * @param fmc FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int fmc_dio_probe(struct fmc_device *fmc)
{
	struct fmc_dio *dio;
	int ret = 0;
	
	/* Driver data */
	dio = devm_kzalloc(&fmc->dev, sizeof(*dio), GFP_KERNEL);
	if (!dio)
		return -ENOMEM;
	
	fmc_set_drvdata(fmc, dio);
	dio->fmc = fmc;
	
	/* FIXME: Recover the DIO internal structure pointer for a global temporary var */
	last_dio = dio;
	
	/* Now register the platform_driver */
	ret = platform_driver_register(&fmc_dio_pl_drv);
	if (ret < 0) {
		fmc_driver_unregister(&fmc_dio_drv);
	}

	return ret;
}

/**
 * FMC DIO remove function
 *
 * This function releases the FMC DIO data structures.
 *
 * @param fmc FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int fmc_dio_remove(struct fmc_device *fmc)
{
	//struct fmc_dio *dio = fmc_get_drvdata(fmc);

	/* Free FMC DIO internal structure */
	/* NOTE: Here is not necessary to free the memory because
	 * we use devm_kmalloc in the probe function
	 */
	 
	platform_driver_unregister(&fmc_dio_pl_drv);

	return 0;
}

/**
 * FMC DIO init function
 *
 * This function is the entry point for the FMC DIO driver.
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_init(void)
{
	int ret;

	ret = fmc_driver_register(&fmc_dio_drv);

	return ret;
}

/**
 * FMC DIO init function
 *
 * This function is the exit point for the FMC DIO driver.
 *
 */
static void fmc_dio_exit(void)
{
	fmc_driver_unregister(&fmc_dio_drv);
}

module_init(fmc_dio_init);
module_exit(fmc_dio_exit);

/* If no gpio lib is there, this weak applies */

/**
 * FMC DIO GPIO init function
 *
 * @param dev FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int __weak fmc_dio_gpio_init(struct fmc_dio *dev)
{
	return 0;
}

/**
 * FMC DIO GPIO exit function
 *
 * @param dev FMC DIO device
 */
void __weak fmc_dio_gpio_exit(struct fmc_dio *dev)
{
	return;
}

MODULE_VERSION(GIT_VERSION);
MODULE_LICENSE("GPL");
