/*
 * Copyright (C) 2012 CERN (www.cern.ch)
 *
 * Author: Alessandro Rubini <rubini@gnudd.com>
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

/* Typical use: "wr-dio-agent wr1 /dev/fmc-dio1" */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>
#include <net/if.h>
#include <netpacket/packet.h>

#include "fmc-dio.h"

static char git_version[] = "version: " GIT_VERSION;

#define RULER_PROTO 0x5752 /* WR */

/*
 * Lazily, use global variables, so the code has less parameter passing.
 * Everything in this file is using "agent_" as a prefix, to ease the
 * reader -- and remember that <TAB> aligns at 8 spaces, not 4.
 */
char			*agent_prgname;
int				agent_sock;
char			*agent_ifname;
int				agent_fd_dio;
char			*agent_dio_name;
struct ifreq	agent_ifr;


/* Boring network stuff extracted from main function */
static int agent_open_wr_sock(char *ifname, char *dio_name)
{
	struct sockaddr_ll addr;
	int sock, fd_dio, ifindex;

	sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (sock < 0) {
		fprintf(stderr, "%s: socket(): %s\n",
			agent_prgname, strerror(errno));
		return -1;
	}

	fd_dio = open(dio_name, O_RDWR);
	if(fd_dio < 0) {
		fprintf(stderr, "%s: open(): %s\n",
			agent_prgname, strerror(errno));
		close(sock);
		return -1;
	}

	memset(&agent_ifr, 0, sizeof(agent_ifr));
	strncpy(agent_ifr.ifr_name, ifname, sizeof(agent_ifr.ifr_name));
	if (ioctl(fd_dio, PRIV_MEZZANINE_ID, (unsigned long) dio_name) < 0
	    /* EAGAIN is special: it means we have no ID to check yet */
		&& errno != EAGAIN) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_ID(%s)): %s\n",
			agent_prgname, dio_name, strerror(errno));
		close(sock);
		return -1;
	}

	/* Retieve the interfaceindex */
	if (ioctl(sock, SIOCGIFINDEX, &agent_ifr) < 0) {
		fprintf(stderr, "%s: SIOCGIFINDEX(%s): %s\n", agent_prgname,
				ifname, strerror(errno));
		close(sock);
		return -1;
	}
	ifindex = agent_ifr.ifr_ifindex;

	/* Bind to the interface, so to be able to receive */
	memset(&addr, 0, sizeof(addr));
	addr.sll_family = AF_PACKET;
	addr.sll_protocol = htons(RULER_PROTO);
	addr.sll_ifindex = ifindex;
	addr.sll_pkttype = PACKET_BROADCAST; /* that's what ruler sends */
	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		fprintf(stderr, "%s: bind(%s): %s\n", agent_prgname,
				ifname, strerror(errno));
		close(sock);
		return -1;
	}

	/* save in globals for later */
	agent_sock = sock;
	agent_ifname = ifname;
	agent_fd_dio = fd_dio;
	agent_dio_name = dio_name;

	return 0;
}

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
}

/* And a simple main with the loop inside */
int main(int argc, char **argv)
{
	int len;
	/* We are receiving stuff in this frame */
	static struct frame {
		struct ether_header h;
		unsigned char pad[2];
		struct wr_dio_cmd cmd;
	} f;

	if ((argc == 2) && (!strcmp(argv[1], "-V"))) {
		print_version(argv[0]);
		exit(0);
	}

	if (argc != 3) {
		fprintf(stderr, "%s: Use \"%s [-V] <wr-if> <dio-dev>\"\n",
			argv[0], argv[0]);
		exit(1);
	}
	agent_prgname = argv[0];

	/* All functions print error messages by themselves, so just exit */
	if (agent_open_wr_sock(argv[1],argv[2]) < 0)
		exit(1);

	while (1) {
		len = recv(agent_sock, &f, sizeof(f), MSG_TRUNC);
		if (len != sizeof(f)) {
			fprintf(stderr, "%s: recevied unexpected frame length"
				" (%i instead of %zu)\n", agent_prgname, len,
				sizeof(f));
			continue;
		}
		if (ntohs(f.h.ether_type) != RULER_PROTO) {
			fprintf(stderr, "%s: received unexpected eth type"
				" (%04x instead of %04x)\n", agent_prgname,
				ntohs(f.h.ether_type), RULER_PROTO);
			continue;
		}

		if (0)
			printf("command %i, ch %i, t %li.%09li\n",
			       f.cmd.command, f.cmd.channel, f.cmd.t[0].tv_sec,
			       f.cmd.t[0].tv_nsec);

		/* Then simply pass it to the hardware */
		if (ioctl(agent_fd_dio, PRIV_MEZZANINE_CMD, (unsigned long)&f.cmd) < 0) {
			fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): "
				"%s\n",	agent_prgname, agent_dio_name,
				strerror(errno));
			return -1;
		}
	}
	exit(0);
}
