/*
 * Copyright (C) 2012 CERN (www.cern.ch)
 *
 * Author: Alessandro Rubini <rubini@gnudd.com>
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "fmc-dio.h"

static char git_version[] = "version: " GIT_VERSION;

char *prgname;
char c;
int fd;
char *devname;

// FIXME: Remove it if not used
void *last_cmd;

struct wr_dio_cmd _cmd;
struct wr_dio_cmd *cmd = &_cmd;

static int parse_ts(char *s, struct timespec *ts)
{
	int i, n;
	unsigned long nano;
	char c;

	/*
 	 * Hairy: if we scan "%ld%lf", the 0.009999 will become 9998 micro.
	 * Thus, scan as integer and string, so we can count leading zeros
	 */

	nano = 0;
	ts->tv_sec = 0;
	ts->tv_nsec = 0;

	if ( (i = sscanf(s, "%ld.%ld%c", &ts->tv_sec, &nano, &c)) == 1)
		return 0; /* seconds only */
	if (i == 3)
		return -1; /* trailing crap */
	if (i == 0)
		if (sscanf(s, ".%ld%c", &nano, &c) != 1)
			return -1; /* leading or trailing crap */

	s = strchr(s, '.') + 1;
	n = strlen(s);
	if (n > 9)
		return -1; /* too many decimals */
	while (n < 9) {
		nano *= 10;
		n++;
	}
	ts->tv_nsec = nano;
	return 0;
}

static int scan_pulse(int argc, char **argv)
{
	char c;

	if (argc != 4 && argc != 6) {
		fprintf(stderr, "%s: %s: wrong number of arguments\n",
			prgname, argv[0]);
		fprintf(stderr, "  Use: %s <channel> <duration> <when> "
			"[<period> <count>]\n", argv[0]);
		return -1;
	}
	if (sscanf(argv[1], "%hi%c", &cmd->channel, &c) != 1
		|| cmd->channel > 4) {
		fprintf(stderr, "%s: %s: not a channel number \"%s\"\n",
			prgname, argv[0], argv[1]);
		return -1;
	}

	/* Duration is first time argument but position 1 for ioctl */
	if (parse_ts(argv[2], cmd->t + 1) < 0) {
		fprintf(stderr, "%s: %s: invalid time \"%s\"\n",
			prgname, argv[0], argv[2]);
		return -1;
	}
	if (cmd->t[1].tv_sec || !cmd->t[1].tv_nsec) {
		fprintf(stderr, "%s: %s: duration must be < 1s and > 0s (got \"%s\")\n",
			prgname, argv[0], argv[2]);
		return -1;
	}
	

	/* Next argument is the "when", position 0 in ioctl timestamp array */
	if (!strcmp(argv[3], "now")) {
		cmd->flags |= WR_DIO_F_NOW;
	} else {
		char *s2 = argv[3];

		if (s2[0] == '+') {
			cmd->flags |= WR_DIO_F_REL;
			s2++;
		}

		if (parse_ts(s2, cmd->t) < 0) {
			fprintf(stderr, "%s: %s: invalid time \"%s\"\n",
				prgname, argv[0], argv[3]);
			return -1;
		}
	}

	/* If argc is 6, we have period and count */
	if (argc == 6) {
		cmd->flags |= WR_DIO_F_LOOP;

		if (parse_ts(argv[4], cmd->t + 2) < 0) {
			fprintf(stderr, "%s: %s: invalid time \"%s\"\n",
				prgname, argv[0], argv[4]);
			return -1;
		}
		if (cmd->t[2].tv_nsec < 1000000)
		{
			fprintf(stderr, "%s: %s: Frequency is greater than 1 kHz or higher than 1 s\n",
				prgname, argv[0]);
			return -1;
		}
		if(cmd->t[2].tv_nsec % 8)
		{
			fprintf(stderr, "%s: %s: Period value is not a multiple of reference frequency \n",
				prgname, argv[0]);
			return -1;
		}
		if (cmd->t[1].tv_nsec >= cmd->t[2].tv_nsec)
		{
			fprintf(stderr, "%s: %s: Pulse length can not be greater than period \n",
				prgname, argv[0]);
			return -1;
		}
		if (sscanf(argv[5], "%i%c", &cmd->value, &c) != 1) {
			fprintf(stderr, "%s: %s: invalid count \"%s\"\n",
				prgname, argv[0], argv[5]);
			return -1;
		}
	}

	if (ioctl(fd, PRIV_MEZZANINE_CMD, (unsigned long) cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s\n",
			prgname, devname, strerror(errno));
			return -1;
	}
	return 0;
}

static int scan_width(int argc, char **argv)
{
	char c;

	if (argc != 3) {
		fprintf(stderr, "%s: %s: wrong number of arguments\n",
			prgname, argv[0]);
		fprintf(stderr, "Use: %s <channel> <duration> \n", argv[0]);
		return -1;
	}
	if (sscanf(argv[1], "%hi%c", &cmd->channel, &c) != 1
		|| cmd->channel > 4) {
		fprintf(stderr, "%s: %s: not a channel number \"%s\"\n",
			prgname, argv[0], argv[1]);
		return -1;
	}

	/* Duration is first time argument but position 1 for ioctl */
	if (parse_ts(argv[2], cmd->t + 1) < 0) {
		fprintf(stderr, "%s: %s: invalid time \"%s\"\n",
			prgname, argv[0], argv[2]);
		return -1;
	}
	if (cmd->t[1].tv_sec || !cmd->t[1].tv_nsec) {
		fprintf(stderr, "%s: %s: duration must be < 1s and > 0s (got \"%s\")\n",
			prgname, argv[0], argv[2]);
		return -1;
	}

	if (ioctl(fd, PRIV_MEZZANINE_CMD, (unsigned long) cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s.\n"
		"The new pulse width must be coherent with the previously programmed period\n",
			prgname, devname, strerror(errno));
			return -1;
	}
	return 0;
}

static int scan_stamp(int argc, char **argv, int ismask)
{
	int i, ch;
	char c;

	cmd->flags = 0;
	if (argc == 3 && !strcmp(argv[2], "wait")) {
		cmd->flags = WR_DIO_F_WAIT;
		argc = 2;
	}
	fprintf(stderr,"scan stamp argc %d\n", argc);
	if (argc == 1) {
		ismask = 1;
		ch = 0x3f;
	} else if (argc == 2) {
		if (sscanf(argv[1], "%i%c", &ch, &c) != 1) {
			fprintf(stderr, "%s: %s: not a channel \"%s\"\n",
				prgname, argv[0], argv[1]);
			exit(1);
		}
		if (ch < 0 || ch > 31 || (!ismask && ch > 5)) {
			fprintf(stderr, "%s: %s: out of range channel \"%s\"\n",
				prgname, argv[0], argv[1]);
			exit(1);
		}
	} else {
		fprintf(stderr, "%s: %s: wrong number of arguments\n",
			prgname, argv[0]);
		if (ismask)
			fprintf(stderr, "  Use: %s [<channel-mask>]\n",
				argv[0]);
		else
			fprintf(stderr, "  Use: %s [<channel>] [wait]\n",
				argv[0]);
		return -1;
	}
	if (ismask)
		cmd->flags = WR_DIO_F_MASK;

	while (1) {
		cmd->channel = ch;
		errno = 0;
		if (ioctl(fd, PRIV_MEZZANINE_CMD, (unsigned long) cmd) < 0 ) {
			if (errno == EAGAIN)
				break;
			fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): "
				"%s\n", prgname, devname, strerror(errno));
		return -1;
		}
		for (i = 0; i < cmd->nstamp; i++)
			printf("ch %i, %9li.%09li (leap second: value=%d, valid=%01x, flag59=%01x, flag61=%01x) \n", cmd->channel,
			       (long)cmd->t[i].tv_sec, cmd->t[i].tv_nsec, cmd->wr_ts_info_buf[i].leap_second, cmd->wr_ts_info_buf[i].leap_second_valid, 
				   cmd->wr_ts_info_buf[i].flag59, cmd->wr_ts_info_buf[i].flag61);
	}
	return 0;
}

static int one_mode(int c, int index)
{
	if (c == '-')
		return 0;
	cmd->channel |= 1 << index;

	//Add error message for channel 0 
	if(index==0 && strchr("dD01",c))
	{
		fprintf(stderr, "Error: Only p/P modes are available as ouput mode for channel 0\n");
		return -1;
	}
	
	switch(c) {
	case 'D':
		cmd->value |= WR_DIO_INOUT_TERM << index;
	case 'd':
		cmd->value |= WR_DIO_INOUT_DIO << index;
		cmd->value |= WR_DIO_INOUT_OUTPUT << index;
		break;

	case 'C':
		cmd->value |= WR_DIO_INOUT_TERM << index;
	case 'c':
		cmd->value |= WR_DIO_INOUT_DIO << index;
		cmd->value |= WR_DIO_INOUT_VALUE << index;
		if(index!=4)
			fprintf(stdout, "Warning: Clock mode is only available for last channel (ch4)\n,"
			 "(on other channel it corresponds to input mode without interruptions)\n");
		break;

	case 'P':
		cmd->value |= WR_DIO_INOUT_TERM << index;
	case 'p':
		cmd->value |= WR_DIO_INOUT_DIO << index;
		cmd->value |= WR_DIO_INOUT_VALUE << index;
		cmd->value |= WR_DIO_INOUT_OUTPUT << index;
		break;

	case 'I':
		cmd->value |= WR_DIO_INOUT_TERM << index;
	case 'i':
		break;

	case '1':
		cmd->value |= WR_DIO_INOUT_VALUE << index;
	case '0':
		cmd->value |= WR_DIO_INOUT_OUTPUT << index;
		break;

	default:
		fprintf(stderr, "%s: mode: invalid mode '%c'\n",
			prgname, c);
		return -1;
	}
	return 0;
}


static int scan_inout(int argc, char **argv)
{
	int i, ch;
	char c;

	cmd->flags = WR_DIO_F_MASK;
	cmd->channel = 0;
	cmd->value = 0;

	if (argc == 2) {
		if (strlen(argv[1]) != 5) {
			fprintf(stderr, "%s: %s: wrong argument \"%s\"\n",
				prgname, argv[0], argv[1]);
			exit(1);
		}
		for (i = 0; i < 5; i++)
			if (one_mode(argv[1][i], i) < 0)
				return -1;
	} else {
		if (argc < 3 || argc > 11 || ((argc & 1) == 0)) {
			fprintf(stderr, "%s: %s: wrong number of arguments\n",
				prgname, argv[0]);
			return -1;
		}
		while (argc >= 3) {
			if (sscanf(argv[1], "%i%c", &ch, &c) != 1
			    || ch < 0 || ch > 4) {
				fprintf(stderr, "%s: mode: invalid channel "
					"\"%s\"\n", prgname,  argv[1]);
				return -1;
			}
			if (strlen(argv[2]) != 1) {
				fprintf(stderr, "%s: mode: invalid mode "
					"\"%s\"\n", prgname,  argv[2]);
				return -1;
			}
			if (one_mode(argv[2][0], ch) < 0)
				return -1;
			argv += 2;
			argc -= 2;
		}
	}
	if (ioctl(fd, PRIV_MEZZANINE_CMD, cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s\n",
			prgname, devname, strerror(errno));
			return -1;
	}
	return 0;
}

static int scan_mask_irq(int argc, char **argv)
{
	int i, ch;
	char c;
	uint32_t value_internal;

	cmd->flags = WR_DIO_F_MASK;
	cmd->channel = 0;
	cmd->value = 0;

	if (argc == 2) {

		cmd->value = WR_DIO_F_MASK_READ_IRQ;
		

	} else if (argc == 3){
		if (argv[2][0] == 'Y') {
			cmd->value = WR_DIO_F_MASK_ENABLE_IRQ;
		}
		else if (argv[2][0] == 'y') {
			cmd->value = WR_DIO_F_MASK_DISABLE_IRQ;
		}
		else
			fprintf(stderr, "%s: mask_irq: invalid argument "
				"\"%s\"\n", prgname,  argv[2]);
	} else {
		fprintf(stderr, "%s: %s: wrong number of arguments\n",
			prgname, argv[0]);
		fprintf(stderr, "  Use: %s <channel> "
			"[<Y/y>]\n", argv[0]);
		return -1;
	}
	
	if (sscanf(argv[1], "%i", &ch) != 1 || ch < 0 || ch > 5) {
		fprintf(stderr, "%s: mask_irq: invalid channel "
			"\"%s\"\n", prgname,  argv[1]);
		return -1;
	}
	cmd->channel |= 1 << ch;

	value_internal = cmd->value; //Save value to read mask

	if (ioctl(fd, PRIV_MEZZANINE_CMD, cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s\n",
			prgname, devname, strerror(errno));
			return -1;
	}
	
	if ( value_internal == WR_DIO_F_MASK_READ_IRQ ) {
		if(cmd->value == WR_DIO_F_MASK_ENABLE_IRQ)
			printf("Ch %i interrupt enabled\n",ch);
		else
			printf("Ch %i interrupt disabled\n",ch);
	}
	
	return 0;
}

static int scan_irq(int argc, char **argv)
{
	char c;

	if (argc != 2 && argc != 4) {
		fprintf(stderr, "%s: %s: wrong number of arguments\n",
			prgname, argv[0]);
		fprintf(stderr, "  Use: %s <when> "
			"[<period> <count>]\n", argv[0]);
		return -1;
	}

	/* Next argument is the "when", position 0 in ioctl timestamp array */
	if (!strcmp(argv[1], "now")) {
		cmd->flags |= WR_DIO_F_NOW;
	} else {
		char *s2 = argv[1];

		if (s2[0] == '+') {
			cmd->flags |= WR_DIO_F_REL;
			s2++;
		}

		if (parse_ts(s2, cmd->t) < 0) {
			fprintf(stderr, "%s: invalid time \"%s\"\n",
				prgname, argv[1]);
			return -1;
		}
	}

	/* If argc is 4, we have period and count */
	if (argc == 4) {
		cmd->flags |= WR_DIO_F_LOOP;

		if (parse_ts(argv[2], cmd->t + 1) < 0) {
			fprintf(stderr, "%s: %s: invalid time \"%s\"\n",
				prgname, argv[0], argv[1]);
			return -1;
		}
		if(cmd->t[1].tv_nsec < 1000000 || cmd->t[1].tv_sec > 2)
		{
			fprintf(stderr, "%s: %s: Frequency is greater than 1 kHz and lower than 0.5 Hz\n",
				prgname, argv[0]);
			return -1;
		}
		if(cmd->t[1].tv_nsec % 8)
		{
			fprintf(stderr, "%s: %s: Period value is not a multiple of reference frequency \n",
				prgname, argv[0]);
			return -1;
		}
		if (sscanf(argv[3], "%i%c", &cmd->value, &c) != 1) {
			fprintf(stderr, "%s: %s: invalid count \"%s\"\n",
				prgname, argv[0], argv[3]);
			return -1;
		}
	}

	if (ioctl(fd, PRIV_MEZZANINE_CMD, (unsigned long) cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s\n",
			prgname, devname, strerror(errno));
			return -1;
	}
	return 0;
}

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
}

int main(int argc, char **argv)
{

	prgname = argv[0];
	argv++, argc--;

	if ((argc == 2) && (!strcmp(argv[1], "-V"))) {
		print_version(argv[0]);
		exit(0);
	}

	if (argc < 2) {
		fprintf(stderr, "%s: use \"%s [-V] <dev> <cmd> [...]\"\n",
			prgname, prgname);
		exit(1);
	}

	devname = argv[0];
	argv++, argc--;

	fd = open(devname,O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "%s: open(): %s\n",
			prgname, strerror(errno));
		exit(1);
	}

	if (ioctl(fd, PRIV_MEZZANINE_ID, (unsigned long) devname) < 0
	    /* EAGAIN is special: it means we have no ID to check yet */
		&& errno != EAGAIN) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_ID(%s)): %s\n",
			prgname, devname, strerror(errno));
	}

	/*
	 * Parse the command line:
	 *
	 * pulse <ch> .<len> <seconds>.<fraction>
	 * pulse <ch> .<len> now
	 * pulse <ch> .<len> +<seconds>.<fraction>
	 *
	 * stamp [<channel>]
	 * stampm [<mask>]
	 *
	 * mode <01234>
	 * mode <ch> <mode> [...]
	 * 
	 * irq <seconds>.<fraction> [<period> <count>]
	 */
	if (!strcmp(argv[0], "pulse")) {
		cmd->command = WR_DIO_CMD_PULSE;
		if (scan_pulse(argc, argv) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "stamp")) {
		cmd->command = WR_DIO_CMD_STAMP;
		fprintf(stderr,"Calling scan_stamp\n");
		if (scan_stamp(argc, argv, 0 /* no mask */) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "stampm")) {
		cmd->command = WR_DIO_CMD_STAMP;
		if (scan_stamp(argc, argv, 1 /* mask */) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "mode")) {
		cmd->command = WR_DIO_CMD_INOUT;
		if (scan_inout(argc, argv) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "irq")) {
		cmd->command = WR_DIO_CMD_IRQ;
		if (scan_irq(argc, argv) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "update_width")) {
		cmd->command = WR_DIO_CMD_UPDATE_WIDTH;
		if (scan_width(argc, argv) < 0)
			exit(1);
	} else if (!strcmp(argv[0], "mask_irq")) {
		cmd->command = WR_DIO_CMD_MASK_IRQ;
		if (scan_mask_irq(argc, argv) < 0)
			exit(1);
	} else {
		fprintf(stderr, "%s: unknown command \"%s\"\n", prgname,
			argv[0]);
		exit(1);
	}

	// FIXME: Is this useful?
	last_cmd = (void *)cmd;
	exit(0);
}
