/*
 * Copyright (C) 2012 CERN (www.cern.ch)
 *
 * Author: Alessandro Rubini <rubini@gnudd.com>
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "fmc-dio.h"

/**
 * This takes two arguments: interface name and channel number
 *
 * This simple tools just show an example of how to program with wr-nic/dio.
 * If you want to measure WR timing we suggest to use the hardwire PPS on channel 0
 * that gives a more accurate precision.
 **/

static char git_version[] = "version: " GIT_VERSION;

static void print_version(char *pname)
{
	printf("%s %s\n", pname, git_version);
}

int main(int argc, char **argv)
{
	struct wr_dio_cmd _cmd;
	struct wr_dio_cmd *cmd = &_cmd;
	char *prgname = argv[0];
	char *devname = "fmc-dio0";
	int fd, ch, charg = 1;
	char c;

	if ((argc == 2) && (!strcmp(argv[1], "-V"))) {
		print_version(argv[0]);
		exit(0);
	}

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "%s: Use \"%s [<devname>] <channel>\"\n",
			prgname, prgname);
		exit(1);
	}
	if (argc == 3) {
		devname = argv[1];
		charg++;
	}
	if (sscanf(argv[charg], "%i%c", &ch, &c) != 1) {
		fprintf(stderr, "%s: Not a channel number \"%s\"\n",
			prgname, argv[charg]);
		exit(1);
	}
	if (ch < 1 || ch > 4) {
		fprintf(stderr, "%s: Out of range channel number \"%s\"\n",
			prgname, argv[charg]);
		exit(1);
	}

	fprintf(stderr, "%s: Using dev \"%s\" and channel %i for "
		"pps output\n", prgname,devname, ch);


	/* The following is standard stuff to access wr-dio */

	fd = open(devname, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "%s: open(): %s\n",
			prgname, strerror(errno));
		exit(1);
	}

	if (ioctl(fd, PRIV_MEZZANINE_ID, (unsigned long) devname) < 0
	    /* EAGAIN is special: it means we have no ID to check yet */
		&& errno != EAGAIN) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_ID(%s)): %s\n",
			prgname, devname, strerror(errno));
	}


	/* Fill the command structure */

	memset(cmd, 0, sizeof(*cmd));

	cmd->command = WR_DIO_CMD_PULSE;
	cmd->channel = ch;
	cmd->flags = WR_DIO_F_REL | WR_DIO_F_LOOP;
	/* Number of loops: -1 <=> Inf */
	cmd->value = -1;
	/* 2s delay to have time to send and process this command */
	cmd->t[0].tv_sec = 2;
	/* 1ms pulse width */
	cmd->t[1].tv_nsec = 1000 * 1000;
	/* Loop period */
	cmd->t[2].tv_sec = 1;


	/* Give it to the driver and we are done */

	if (ioctl(fd, PRIV_MEZZANINE_CMD, (unsigned long) cmd) < 0) {
		fprintf(stderr, "%s: ioctl(PRIV_MEZZANINE_CMD(%s)): %s\n",
			prgname, devname, strerror(errno));
		exit(1);
	}

	exit(0);
}
